#include <QCoreApplication>
#include <QTextStream>
#include <QStorageInfo>
#include <QUuid>
#include <QProcess>

#include <botan-2/botan/rng.h>
#include <botan-2/botan/auto_rng.h>
#include <botan-2/botan/cipher_mode.h>
#include <botan-2/botan/hex.h>

#include "dotenv.h"
#include "parentimage.h"
#include "extpartition.h"

int main(int argc, char *argv[])
{
    QCoreApplication a(argc, argv);

    QTextStream in(stdin);
    QTextStream out(stdout);

    // Passwords / Private Keys fetch from environment variables file
    dotenv::env.load_dotenv(".env", true);
    const QString rootPassword = QString::fromStdString((std::string) dotenv::env["SU_PASSWORD"]);
    const QString imageFilePath = QString::fromStdString((std::string) dotenv::env["PARTITION_IMAGE_PATH"]);
    const QString aesKey = QString::fromStdString((std::string) dotenv::env["PARTITION_AES_KEY"]);
    const QString sshPrivKeyPath = QString::fromStdString((std::string) dotenv::env["SERVER_SSH_PRIV_KEY_PATH"]);
    const QString userId = QString::fromStdString((std::string) dotenv::env["USER_ID"]);
    const QString userApiKey = QString::fromStdString((std::string) dotenv::env["USER_PRIVATE_API_KEY"]);

    //////
    // Use Case: Partition encryption & Upload to server
    //////

    // List mounted partitions
    QList<Partition *> mountedPartitions;
    QList<QStorageInfo> mountedPartitionsInfo = QStorageInfo::mountedVolumes();
    QMutableListIterator<QStorageInfo> i(mountedPartitionsInfo);
    while (i.hasNext()) {
        const QStorageInfo partition = i.next();
        if (partition.isValid() && partition.isReady()) {
            if (partition.fileSystemType() == "ext4") {
                ExtPartition *currentPartition = new ExtPartition();
                currentPartition->setId(QUuid::createUuid().toString());
                currentPartition->setLabel(partition.displayName());
                currentPartition->setDevice(partition.device());
                currentPartition->setSize(partition.bytesTotal());
                currentPartition->setRootPath(partition.rootPath());
                mountedPartitions.push_back(currentPartition);
            } else {
                i.remove();
            }
        }
    }

    out << "Choose a partition to backup:" << endl;
    for (int i = 0; i < mountedPartitions.length(); ++i) {
        const ExtPartition *extPartition = dynamic_cast<ExtPartition *>(mountedPartitions[i]);
        if (extPartition) {
            out << (i + 1) << " - " << extPartition->getDevice() << " (" << extPartition->getLabel() << ") " << endl;
        }
    }
    out << "0 - Exit program " << endl;

    int inputPartition;
    QString inputStr;
    while (in.readLineInto(&inputStr)) {
        inputPartition = inputStr.toInt();
        if (inputPartition > 0) {
            break;
        } else if (inputPartition == 0) {
            std::exit(1);
        }
    }
    if (inputPartition > 0) {
        ExtPartition* extPartition = dynamic_cast<ExtPartition*>(mountedPartitions[inputPartition - 1]);
        if (extPartition) {
            // Unmount the partition
            extPartition->unmount(rootPassword, &out);
            // Create partition image
            ParentImage* image = dynamic_cast<ParentImage*>(extPartition->createImage(rootPassword, imageFilePath, &out));
            // Mount the partition again
            extPartition->mount(rootPassword, extPartition->getRootPath(), &out);
            // Give public privileges to the image
            image->setPrivileges(rootPassword, 777, &out);
            // File Encryption
            out << "> Encrypting image.." << endl;
            qint64 bytesWritten = image->encrypt(aesKey);
            QString initVector = image->getEncryptedInitVector();
            out << ">> Encrypted successfully!" << endl;
            out << ">>> Encrypted unique key: " << image->getEncryptedAesKey() << endl;
            out << ">>> Encrypted IV: " << image->getEncryptedInitVector() << endl;
            out << ">>> " << bytesWritten << " bytes written.\n" << endl;

//            // File Decryption (not part of this use case)
//            bytesWritten = image->decrypt(image->getAesKey(aesKey), image->getInitVector(aesKey));
//            out << ">>> Decrypted successfully!" << endl;
//            out << ">> " << bytesWritten << " bytes written." << endl;

            // File Upload
            image->uploadToServer(sshPrivKeyPath, userId, userApiKey, &out);
            out << ">> Encrypted image uploaded successfully!" << endl;
            image->deleteLater();
        }
    }

    qDeleteAll(mountedPartitions);
    mountedPartitions.clear();
    out.flush();
    std::exit(1);
    return a.exec();
}
