#include "partitionimage.h"

#include <QUuid>
#include <QProcess>
#include <QEventLoop>
#include <QFile>
#include <QtNetwork/QNetworkAccessManager>
#include <QtNetwork/QNetworkReply>
#include <QJsonDocument>
#include <QJsonArray>
#include <QJsonObject>
#include <QJsonValue>

#include <botan-2/botan/rng.h>
#include <botan-2/botan/auto_rng.h>
#include <botan-2/botan/cipher_mode.h>
#include <botan-2/botan/hex.h>

// Constructor
PartitionImage::PartitionImage(const QString &uniqueId, const QString &inputPath, const QString &inputFileExtension, QObject *parent) : QObject(parent)
{
    id = uniqueId;
    path = inputPath + id + "." + inputFileExtension;
    imageFile.setFileName(path);
    fileExtension = inputFileExtension;
}

// Getters & Setters
QString PartitionImage::getId() const
{
    return id;
}

void PartitionImage::setId(const QString &value)
{
    id = value;
}

QString PartitionImage::getParentId() const
{
    return parentId;
}

void PartitionImage::setParentId(const QString &value)
{
    parentId = value;
}

QString PartitionImage::getLabel() const
{
    return label;
}

void PartitionImage::setLabel(const QString &value)
{
    label = value;
}

QString PartitionImage::getPath() const
{
    return path;
}

void PartitionImage::setPath(const QString &value)
{
    path = value;
}

QString PartitionImage::getFileExtension() const
{
    return fileExtension;
}

void PartitionImage::setFileExtension(const QString &value)
{
    fileExtension = value;
}

int PartitionImage::getSize() const
{
    return size;
}

void PartitionImage::setSize(int value)
{
    size = value;
}

QDateTime PartitionImage::getCreatedAt() const
{
    return createdAt;
}

void PartitionImage::setCreatedAt(const QDateTime &value)
{
    createdAt = value;
}

QString PartitionImage::getEncryptedAesKey() const
{
    return encryptedAesKey;
}

void PartitionImage::setEncryptedAesKey(const QString &value, const bool &isEncrypted, const QString &masterAesKey)
{
    if (isEncrypted) {
        encryptedAesKey = value;
    } else {
        Botan::AutoSeeded_RNG rng;

        // Set cypher mode to AES-128/CBC/PKCS7
        std::unique_ptr<Botan::Cipher_Mode> enc = Botan::Cipher_Mode::create("AES-256/CBC/PKCS7", Botan::ENCRYPTION);
        // Set master key as encryption key
        enc->set_key(Botan::hex_decode(masterAesKey.toStdString()));
        // Generate, store and set random nonce (Initialization Vector)
        Botan::secure_vector<uint8_t> iv = rng.random_vec(enc->default_nonce_length());
        aesKeyEncryptionIv = QString::fromStdString(Botan::hex_encode(iv));
        enc->start(iv);
        // Execute AES-128/CBC/PKCS7 on the unique key
        const std::string strValue = value.toStdString();
        Botan::secure_vector<uint8_t> securizedValue(strValue.data(), strValue.data() + strValue.length());
        enc->finish(securizedValue);

        encryptedAesKey = QString::fromStdString(Botan::hex_encode(securizedValue));
    }
}

QString PartitionImage::getEncryptedInitVector() const
{
    return encryptedInitVector;
}

void PartitionImage::setEncryptedInitVector(const QString &value, const bool &isEncrypted, const QString &masterAesKey)
{
    if (isEncrypted) {
        encryptedInitVector = value;
    } else {
        Botan::AutoSeeded_RNG rng;

        // Set cypher mode to AES-128/CBC/PKCS7
        std::unique_ptr<Botan::Cipher_Mode> enc = Botan::Cipher_Mode::create("AES-256/CBC/PKCS7", Botan::ENCRYPTION);
        // Set master key as encryption key
        enc->set_key(Botan::hex_decode(masterAesKey.toStdString()));
        // Generate, store and set random nonce (Initialization Vector)
        Botan::secure_vector<uint8_t> iv = rng.random_vec(enc->default_nonce_length());
        initVectorEncryptionIv = QString::fromStdString(Botan::hex_encode(iv));
        enc->start(iv);
        // Execute AES-128/CBC/PKCS7 on the unique key
        const std::string strValue = value.toStdString();
        Botan::secure_vector<uint8_t> securizedValue(strValue.data(), strValue.data() + strValue.length());
        enc->finish(securizedValue);

        encryptedInitVector = QString::fromStdString(Botan::hex_encode(securizedValue));
    }
}


// Public functions
qint64 PartitionImage::encrypt(const QString &masterAesKey)
{
    Botan::AutoSeeded_RNG rng;
    QTextStream debug(stdout);//


    // Open partition file for read
    if (!imageFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        throw QString("Failed to open file").toUtf8();
    };

    // Set cypher mode to AES-128/XTS
    std::unique_ptr<Botan::Cipher_Mode> enc = Botan::Cipher_Mode::create("AES-128/XTS", Botan::ENCRYPTION);

    // Set given encryption key
    // Generate, store and set random AES encryption key
    Botan::secure_vector<uint8_t> aesKey = rng.random_vec(enc->maximum_keylength());
    debug << "[DEBUG] Key: " << QString::fromStdString(Botan::hex_encode(aesKey)) << endl;//
    this->setEncryptedAesKey(QString::fromStdString(Botan::hex_encode(aesKey)), false, masterAesKey);
    enc->set_key(aesKey);

    // Generate, store and set random nonce (Initialization Vector)
    Botan::secure_vector<uint8_t> initVector = rng.random_vec(enc->default_nonce_length());
    QString initVectorStr = QString::fromStdString(Botan::hex_encode(initVector));
    debug << "[DEBUG] IV: " << initVectorStr << endl;//
    this->setEncryptedInitVector(initVectorStr, false, masterAesKey);
    enc->start(initVector);

    // Copy input data to a buffer that will be encrypted
    QByteArray fileContents = imageFile.readAll();
    const uint8_t *begin = reinterpret_cast<uint8_t *>(fileContents.data());
    const uint8_t *end = begin + fileContents.length();
    Botan::secure_vector<uint8_t> securizedFileContents(begin, end);
    imageFile.close();

    // Execute AES-128/XTS on the file contents
    enc->finish(securizedFileContents);

    // Overwrite image file with encrypted version
    imageFile.open(QIODevice::WriteOnly);
    QByteArray encFileContents(reinterpret_cast<const char *>(securizedFileContents.data()), securizedFileContents.size());
    qint64 bytesWritten = imageFile.write(encFileContents);
    imageFile.close();

    // TBD: Update object in DB

    return bytesWritten;
}

qint64 PartitionImage::decrypt(const QString &masterAesKey, const QString &initVector)
{
    Botan::AutoSeeded_RNG rng;

    // Open partition file for read
    if (!imageFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        throw QString("Failed to open file").toUtf8();
    };

    // Set cypher mode to AES-128/XTS
    std::unique_ptr<Botan::Cipher_Mode> enc = Botan::Cipher_Mode::create("AES-128/XTS", Botan::DECRYPTION);

    // Set given encryption key
    const std::vector<uint8_t> aesKeyBytes = Botan::hex_decode(masterAesKey.toStdString());
    enc->set_key(aesKeyBytes);

    // Set stored nonce (Initialization Vector)
    std::vector<uint8_t> initVectorBytes = Botan::hex_decode(initVector.toStdString());
    enc->start(initVectorBytes);

    // Copy input data to a buffer that will be encrypted
    QByteArray fileContents = imageFile.readAll();
    const uint8_t *begin = reinterpret_cast<uint8_t *>(fileContents.data());
    const uint8_t *end = begin + fileContents.length();
    Botan::secure_vector<uint8_t> securizedFileContents(begin, end);
    imageFile.close();

    // Execute AES-128/XTS on the file contents
    enc->finish(securizedFileContents);

    // Overwrite image file with encrypted version
    imageFile.open(QIODevice::WriteOnly);
    QByteArray decFileContents(reinterpret_cast<const char *>(securizedFileContents.data()), securizedFileContents.size());
    qint64 bytesWritten = imageFile.write(decFileContents);
    imageFile.close();

    // TBD: Update object in DB

    return bytesWritten;
}

int PartitionImage::setPrivileges(const QString &suPassword, int octalPermissions, QTextStream *outputStream)
{
    QProcess *chmod = new QProcess;
    chmod->setProgram("bash");
    chmod->setProcessChannelMode(QProcess::ForwardedChannels);
    chmod->start();
    if(!chmod->waitForStarted()) throw QString(">>> chmod init error!");
    QString chmodSuCommand = QString("echo ") + suPassword + QString(" | ") + QString("sudo -S chmod ") + QString::number(octalPermissions) + " " + this->path + QString(" && exit \n");
    while (!chmod->isOpen());
    chmod->write(chmodSuCommand.toUtf8());
    if (outputStream != nullptr) {
        *outputStream << "\n> Setting image file privileges...\n" << endl;
    }
    if(!chmod->waitForFinished()) throw QString(">>> chmod error");
    chmod->deleteLater();
    return 1;
}

int PartitionImage::uploadToServer(const QString &sshPrivKeyPath, const QString &userId, const QString &apiKey, QTextStream *outputStream) {
    QNetworkAccessManager manager;
    // TBD: Hardcoded for now
    QNetworkRequest request(QUrl("http://acop.io/api/users/" + userId + "/files"));
    request.setRawHeader("Content-Type", "application/json");
    // Prepare request JSON
    QJsonDocument requestJson;
    QJsonObject requestJsonObject;
    requestJsonObject["apiKey"] = apiKey;
    requestJsonObject["fileId"] = this->id;
    requestJsonObject["fileExtension"] = this->fileExtension;
    requestJsonObject["encryptedAesKey"] = this->encryptedAesKey;
    requestJsonObject["encryptedInitVector"] = this->encryptedInitVector;
    requestJson.setObject(requestJsonObject);
    if (outputStream != nullptr) {
        *outputStream << "> Sending API request...\n" << endl;
    }
    QNetworkReply* reply = manager.post(request, requestJson.toJson());
    // Wait for reply
    QEventLoop loop;
    QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
    loop.exec();
    if (outputStream != nullptr && reply->error()) {
        *outputStream << reply->errorString() << endl;
    }
    // Process response JSON
    QJsonDocument responseJson = QJsonDocument::fromJson(reply->readAll());
    if (responseJson.isEmpty()) return 0;
    QJsonObject responseJsonObject = responseJson.object();
    if (responseJsonObject.keys().indexOf("status") == -1 || responseJsonObject["status"].toInt() != 1) return 0;
    if (responseJsonObject.keys().indexOf("uploadPath") == -1) return 0;
    QString uploadPath = responseJsonObject["uploadPath"].toString();
    // Initiate SCP transfer to the upload path
    QProcess *scp = new QProcess;
    scp->setProgram("scp");
    scp->setProcessChannelMode(QProcess::ForwardedChannels);
    // TBD: Hardcoded for now
    scp->setArguments({"-P", "2323", "-i", sshPrivKeyPath, this->path, QString("pi@acop.io") + QString(":") + QString(uploadPath)});
    if (outputStream != nullptr) {
        *outputStream << "> Sending file...\n" << endl;
    }
    scp->start();
    if(!scp->waitForStarted()) throw QString(">>> chmod init error!");
    if(!scp->waitForFinished()) throw QString(">>> chmod error");
    return 1;
}

